/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.core;

// API classes
import com.rempl.api.Artifact;
import com.rempl.api.Loader;
import com.rempl.api.LoadingException;
import com.rempl.api.Logger;
import com.rempl.api.ROM;
import com.rempl.api.ROMConfiguration;
import com.rempl.api.ROMException;
import com.rempl.api.Reader;
import com.rempl.api.RemplFactory;
import com.rempl.api.Rempler;
import com.rempl.api.Reporter;
import com.rempl.api.RuntimeProblem;

/**
 * Factory of {@link Rempler}s, and other classes
 * in the package.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: RemplFactoryImpl.java 776 2011-03-18 22:35:07Z yegor256@yahoo.com $
 */
public final class RemplFactoryImpl implements RemplFactory {

    /**
     * Loader to use for loading artifacts.
     */
    private Loader loader = new RemplFactoryImpl.FalseLoader();

    /**
     * {@inheritDoc}
     */
    @Override
    public void setLoader(final Loader ldr) {
        if (ldr == null) {
            this.loader = new RemplFactoryImpl.FalseLoader();
        } else {
            this.loader = ldr;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Loader getLoader() {
        return this.loader;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Rempler makeRempler(final ROM rom) {
        if (rom == null) {
            throw new IllegalArgumentException("#makeRempler(): is NULL");
        }
        final Rempler rempler = new RemplerImpl();
        rempler.configure(this.load(rempler, rom));
        this.loadReporters(rempler, rom);
        return rempler;
    }

    /**
     * Load ROM into given REMPLER.
     * @param rempler The rempler
     * @param rom The ROM
     * @return The configuration just retrieved
     * @see #makeRempler(ROM)
     */
    private ROMConfiguration load(final Rempler rempler, final ROM rom) {
        try {
            final ROMConfiguration config = new ROMConfiguration(null);
            for (Artifact dependency : rom.dependencies()) {
                if (this.loader == null) {
                    Logger.warn(
                        this,
                        "Ignoring %s since loader is not set",
                        dependency
                    );
                    continue;
                }
                Logger.info(
                    this,
                    "Adding reader '%s' to the Rempler",
                    dependency
                );
                final ROM dep = this.loader.load(dependency);
                // a recursive call to itself
                config.append(this.load(rempler, dep));
                if (dep.hasWorker()) {
                    rempler.append((Reader) dep.getWorker());
                }
            }
            config.append(rom.configuration());
            return config;
        } catch (ROMException ex) {
            throw RuntimeProblem.make(ex);
        } catch (LoadingException ex) {
            throw RuntimeProblem.make(ex);
        }
    }

    /**
     * Load reporters and append them to the Rempler.
     * @param rempler The rempler
     * @param rom The ROM
     * @see #makeRempler(ROM)
     * @todo #16 This is a temporary solution. We need to
     *    use something like this:
     *    dependency.getGroupId() + "/" + dependency.getArtifactId()
     *    We can't do it now, since getGroupId() and other methods
     *    are being introduced by ticket #15, which is still not in
     *    trunk. That's why such an ugly implementation.
     */
    private void loadReporters(final Rempler rempler, final ROM rom) {
        try {
            Integer num = 1;
            for (Artifact dependency : rom.reporters()) {
                Logger.info(
                    this,
                    "Adding reporter '%s' to the Rempler",
                    dependency
                );
                final ROM dep = this.loader.load(dependency);
                rempler.append(
                    (Reporter) dep.getWorker(),
                    "report-" + num
                );
                num += 1;
            }
        } catch (ROMException ex) {
            throw RuntimeProblem.make(ex);
        } catch (LoadingException ex) {
            throw RuntimeProblem.make(ex);
        }
    }

    /**
     * False loader, instead of a real one.
     */
    private static final class FalseLoader implements Loader {
        /**
         * {@inheritDoc}
         */
        @Override
        public ROM load(final Artifact artifact) throws LoadingException {
            try {
                // this will lead to exception in all cases
                return new ROM("Loader Is Not Set");
            } catch (ROMException ex) {
                throw new LoadingException(ex);
            }
        }
        /**
         * {@inheritDoc}
         */
        @Override
        public Class forName(final String name) throws LoadingException {
            try {
                return this.getClass().forName(name);
            } catch (ClassNotFoundException ex) {
                throw new LoadingException(ex);
            }
        }
    }

}
