/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.core;

// API classes
import com.rempl.api.Logger;
import com.rempl.api.Model;
import com.rempl.api.ROMConfiguration;
import com.rempl.api.Reader;
import com.rempl.api.Rempler;
import com.rempl.api.Reporter;

// java input/output
import java.io.File;
import java.io.IOException;

// collections/maps processing
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Builder of a {@link Model}.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: RemplerImpl.java 776 2011-03-18 22:35:07Z yegor256@yahoo.com $
 */
public final class RemplerImpl implements Rempler {

    /**
     * ROM configuration.
     */
    private ROMConfiguration configuration;

    /**
     * Ordered list of {@link Reader}s.
     */
    private List<Reader> readers = new ArrayList<Reader>();

    /**
     * Map of {@link Reporter}s.
     */
    private Map<String, Reporter> reporters
        = new HashMap<String, Reporter>();

    /**
     * {@inheritDoc}
     * @todo #3 duplicate readers have to be ignored or just rejected
     */
    @Override
    public void append(final Reader reader) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#append(reader)");
        }
        if (reader == null) {
            throw new IllegalArgumentException("#append(): reader is NULL");
        }
        this.readers.add(reader);
        reader.setRempler(this);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void append(final Reporter reporter, final String dir) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#append(reporter)");
        }
        if (reporter == null) {
            throw new IllegalArgumentException("#append(): reporter is NULL");
        }
        if (this.reporters.containsKey(dir)) {
            throw new IllegalStateException(
                "Directory " + dir + " already set"
            );
        }
        this.reporters.put(dir, reporter);
        reporter.setRempler(this);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Model read() throws IOException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#read()");
        }
        final Model model = this.emptyModel();
        List<Integer> running = new ArrayList<Integer>();
        for (Integer i = 0; i < this.readers.size(); i += 1) {
            running.add(i);
        }
        Logger.debug(this, "%d readers to be executed", running.size());
        do {
            final List<Integer> next = new ArrayList<Integer>();
            for (Integer i : running) {
                final Reader reader = this.readers.get(i);
                final String name = reader.getClass().getName();
                Logger.debug(this, "reading with %s (#%d)", name, i);
                if (reader.read(model)) {
                    Logger.debug(this, "%s (%d) needs more", name, i);
                    next.add(i);
                } else {
                    Logger.debug(this, "removing %s (%d) from list", name, i);
                }
            }
            running = next;
        } while (running.size() > 0);
        return model;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Model emptyModel() {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#emptyModel()");
        }
        return new SimpleModel();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void report(final Model model, final File dir) throws IOException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#report(model, dir)");
        }
        if (model == null) {
            throw new IllegalArgumentException("#report(): model is NULL");
        }
        if (dir == null) {
            throw new IllegalArgumentException("#report(): dir is NULL");
        }
        if (!dir.exists()) {
            throw new IllegalArgumentException(
                "Directory doesn't exist: " + dir
            );
        }
        for (Map.Entry<String, Reporter> pair : this.reporters.entrySet()) {
            final File subdir = new File(dir.toString() + "/" + pair.getKey());
            if (subdir.mkdirs()) {
                Logger.info(this, "Directory %s created", dir);
            }
            pair.getValue().report(model, subdir);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void configure(final ROMConfiguration config) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#configure(config)");
        }
        this.configuration = config;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public ROMConfiguration configuration() {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#configuration()");
        }
        return this.configuration;
    }

}
