/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.core;

// API classes
import com.rempl.api.Element;
import com.rempl.api.Logger;
import com.rempl.api.Model;
import com.rempl.api.RuntimeProblem;

// ANTLR3-generated parsers and lexers
import com.rempl.core.model.QueryLexer;
import com.rempl.core.model.QueryParser;
import com.rempl.core.model.TransformLexer;
import com.rempl.core.model.TransformParser;

// for collection implementation
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

// ANTLR3 runtime for grammar/lexical processing
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.TokenStream;

// IO utility class
import org.apache.commons.io.IOUtils;

/**
 * MOF 2.0 model abstraction.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: SimpleModel.java 692 2011-03-16 18:07:50Z yegor256@yahoo.com $
 * @todo #3 This implementation is extremely simplified and won't
 *          be able to do anything good in any real-life project.
 *          It has to be extended significantly or replaced at all
 *          by Eclipse EMF components.
 */
public final class SimpleModel implements Model {

    /**
     * List of elements.
     */
    private Collection<Element> elements = new ArrayList<Element>();

    /**
     * {@inheritDoc}
     */
    @Override
    public void transform(final String query) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#transform(query)");
        }
        final CharStream input = new ANTLRStringStream(query);
        final TransformLexer lexer = new TransformLexer(input);
        final TokenStream tokens = new CommonTokenStream(lexer);
        final TransformParser parser = new TransformParser(tokens);
        Collection<Element> found;
        try {
            found = parser.elements();
        } catch (org.antlr.runtime.RecognitionException ex) {
            throw RuntimeProblem.make(ex);
        }
        for (Element e : found) {
            this.elements.add(e);
        }
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#transform(..): %d elements: %s",
                found.size(),
                query
            );
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Collection<Element> query(final String query) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#query(query)");
        }
        final CharStream input = new ANTLRStringStream(query);
        final QueryLexer lexer = new QueryLexer(input);
        final TokenStream tokens = new CommonTokenStream(lexer);
        final QueryParser parser = new QueryParser(tokens);
        String context;
        try {
            context = parser.context();
        } catch (org.antlr.runtime.RecognitionException ex) {
            throw RuntimeProblem.make(ex);
        }
        final Collection<Element> list = new ArrayList<Element>();
        for (Element e : this.elements) {
            if (e.type().equals(context)) {
                list.add(e);
            }
        }
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#query('%s'): found %d elements",
                query,
                list.size()
            );
        }
        return list;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void save(final java.io.Writer writer) throws java.io.IOException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#save(writer)");
        }
        final StringBuffer buffer = new StringBuffer();
        for (Element e : this.elements) {
            if (e.name() != null) {
                buffer.append(e.name());
            }
            buffer.append(":" + e.type() + " { ");
            for (Map.Entry<String, String> pair : e.entrySet()) {
                buffer.append(
                    pair.getKey()
                    + ": \""
                    + pair.getValue()
                    + "\"; "
                );
            }
            buffer.append(" };\n");
        }
        writer.write(buffer.toString());
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#save(%s): saved %d bytes",
                writer.getClass().getName(),
                buffer.length()
            );
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void load(final java.io.Reader reader) throws java.io.IOException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#load(reader)");
        }
        final String xmi = IOUtils.toString(reader);
        this.transform(xmi);
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#load(%s): loaded %d bytes",
                reader.getClass().getName(),
                xmi.length()
            );
        }
    }


}
